#
# Copyright 2018 Fairphone B.V.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""The (interactive) shell interface to Android Maestro."""

import cmd
from contextlib import contextmanager
import logging
from typing import Dict, List, Sequence, Set

from adb.adb_commands import AdbCommands
import adb.usb_exceptions
from android_tools import Device
import android_tools.utils
import usb1

from .loads import Monkey
from .session import Session


LOG = logging.getLogger(__name__)


class DeviceAdbError(Exception):
    """An error wrapper to catch ADB errors on a device.

    Attributes:
        device: The device on which the error happened.
        reason: The reason of the error (human readable).
        root: The wrapped root exception.

    """

    def __init__(self, device: Device, reason: str, root: Exception):
        """Initialise the error wrapper."""
        self.device = device
        self.root = root
        self.reason = reason
        super().__init__(f"Device {device.serial} error: {reason}")


class MaestroShell(cmd.Cmd):
    """Orchestrate system load session on Android devices.

    This interactive shell is able to start many :py:class:`.session.Session` on
    many Android :py:class:`android_tools.Device` and to monitor them:

    - `run [SERIALs...]`: Start a session on connected devices.
    - `session [UUIDs...]`: List the running sessions.

    Session UUID's and device serials are auto-completed.

    """

    _DO_SESSIONS_COLUMNISED_SECTION = "Session {uuid: <36}: {state}"
    _DO_SESSIONS_COLUMNISED_DEVICE = "└── {device: <24} {state: <8}"

    def __init__(self):
        """Initialise the shell."""
        super().__init__()
        # Get an AuthSigner instead of relying on the Android tools' default one
        self._adb_signer = android_tools.utils.auth_signer_from_user_keys()
        self._sessions: Dict[str, Session] = {}
        self.prompt = "maestro> "

    @contextmanager
    def _adb_on_device(
        self, device: Device, exception_as_debug: bool = False
    ) -> AdbCommands:
        """Wrap around the ADB context manager of a device.

        Protect the shell from exceptions that can be raised when communicating
        with a device through ADB -- i.e. any logic using
        :py:func:`android_tools.Device.adb` -- by creating a new context
        manager only raising one type of error (:py:class:`DeviceAdbError`).

        The list of handled exceptions, and the potential triggers, is:

        - :py:class:`adb.usb_exceptions.DeviceAuthError`: When the device has
            not authorised the signing keys used by the host.
        - :py:class:`adb.usb_exceptions.DeviceNotFoundError`: When the device
            is not found on the USB hub.
        - :py:class:`adb.usb_exceptions.ReadFailedError` and
            :py:class:`adb.usb_exceptions.WriteFailedError`: When the device is
            not responding to ADB commands. The device is likely frozen.
        - :py:class:`usb1.USBErrorBusy`: When the USB channel is already busy,
            either because the ADB server is running (incompatible with the
            Python ADB library!) or because more than one command was issued
            to a device at once.

        The exceptions will always be logged, though their severity can be
        dialed down with the `exception_as_debug` argument. It is an intended
        behaviour that exceptions cannot be silenced as they are meaningful
        ("Errors should never pass silently."). The rationale to allow for a
        different level is that during regular interactive sessions, one does
        not usually want too many details. Raising the log level to display
        debug messages still allows to see these internal traces.

        Arguments:
            device: The device to open an ADB channel with.
            exception_as_debug (optional): Whether to log the exceptions with a
                `DEBUG` level or with an `ERROR` level (default).
        Returns:
            The context manager :py:class:`adb.adb_commands.AdbCommands` yielded
            by `device.adb()`.
        Raises:
            DeviceAdbError: If an error is caught and logged, it is re-raised
                with the :py:class:`DeviceAdbError` wrapper.

        """
        _log = LOG.debug if exception_as_debug else LOG.error

        try:
            with device.adb() as _adb:
                yield _adb
        except adb.usb_exceptions.DeviceAuthError as exception:
            _log(device.serial, exc_info=exception)
            raise DeviceAdbError(device, "Host not authorised", exception)
        except adb.usb_exceptions.DeviceNotFoundError as exception:
            _log(device.serial, exc_info=exception)
            raise DeviceAdbError(device, "Not found", exception)
        except (
            adb.usb_exceptions.ReadFailedError,
            adb.usb_exceptions.WriteFailedError,
        ) as exception:
            _log(device.serial, exc_info=exception)
            raise DeviceAdbError(device, "No response", exception)
        except usb1.USBErrorBusy as exception:
            # TODO adb: Wrap the exception in e.g. LibusbWrappingError
            _log(device.serial, exc_info=exception)
            raise DeviceAdbError(device, "USB channel busy", exception)

    @staticmethod
    def _known_items_matching(
        known_items: Sequence[str], prefix: str, line: str
    ) -> List[str]:
        """Return the known items matching an input prefix.

        Helper method for command completion based on a list of known items.
        The items already listed are ignored, as well as those not matching the
        input prefix.

        Arguments:
            known_items: The sequence of known items (lower case).
            prefix: The input prefix (lower case) that should be matched.
            line: The current input line (lower case), potentially containing
                known items.
        Returns:
            The list of sorted matches.

        """
        if not isinstance(known_items, Set):
            known_items = set(known_items)
        listed_items = set(item for item in line.split() if item in known_items)
        remaining_items = known_items - listed_items
        return sorted(
            [item for item in remaining_items if item.startswith(prefix)]
        )

    def do_exit(self, line: str):
        """Exit the shell."""
        # pylint: disable=unused-argument, W0613
        # pylint: disable=no-self-use, R0201
        return True

    do_EOF = do_exit

    def complete_sessions(
        self, text: str, line: str, begidx: int, endidx: int
    ) -> List[str]:
        """Complete the `sessions` command.

        Complete the text to match against known session UUID's. The session
        UUID's already present on the line are removed from the matches.

        Arguments:
            text: The string prefix we are attempting to match.
            line: The current input line with leading whitespace removed.
            begidx: The beginning index of the prefix text.
            endidx: The ending index of the prefix text.
        Returns:
            The list of sorted matches.

        """
        # pylint: disable=unused-argument, W0613
        return self._known_items_matching(
            self._sessions.keys(), text.lower(), line.lower()
        )

    def do_sessions(self, line: str) -> None:
        """Describe the status of known sessions: `sessions [UUID...]`.

        Only look for known sessions which UUID's match the positional
        parameters. With no UUID's specified, look at all known sessions.
        (Known sessions are sessions started from this shell.)

        """
        if line:
            uuids = set(line.split())
            for uuid in uuids - self._sessions.keys():
                print(f"Ignoring unknown session UUID: {uuid}")
            sessions = {
                uuid: self._sessions[uuid]
                for uuid in uuids
                if uuid in self._sessions
            }
        else:
            sessions = self._sessions

        for uuid, session in sessions.items():
            print(
                self._DO_SESSIONS_COLUMNISED_SECTION.format(
                    uuid=uuid, state=session.state.value
                )
            )
            for device in session.devices:
                try:
                    with self._adb_on_device(device, exception_as_debug=True):
                        state = session.get_device_state(device)
                except DeviceAdbError:
                    state = None

                print(
                    self._DO_SESSIONS_COLUMNISED_DEVICE.format(
                        device=str(device), state=state.value if state else "??"
                    )
                )

    def complete_run(
        self, text: str, line: str, begidx: int, endidx: int
    ) -> List[str]:
        """Complete the `run` command.

        Complete the text to match against device serials. The device serials
        already present on the line are removed from the matches.

        Arguments:
            text: The string prefix we are attempting to match
            line: The current input line with leading whitespace removed.
            begidx: The beginning index of the prefix text.
            endidx: The ending index of the prefix text.
        Returns:
            The list of sorted matches.

        """
        # pylint: disable=unused-argument, W0613
        all_serials = set(serial.lower() for serial in Device.list_serials())
        return self._known_items_matching(
            all_serials, text.lower(), line.lower()
        )

    def do_run(self, line: str) -> None:
        """Begin a system load session on devices: `run [SERIAL...]`.

        Initialise a new demo session and start it on all devices reachable by
        ADB which serial numbers match the positional parameters. With no
        SERIALs specified, find and use all devices available.

        """
        serials = line.split() or Device.list_serials()
        devices = []

        for serial in serials:
            device = Device(serial, self._adb_signer)
            try:
                with self._adb_on_device(device, exception_as_debug=True):
                    devices.append(device)
                    print(f"Device {device} added to the pool.")
            except DeviceAdbError as exception:
                print(
                    f"Could not add device {device.serial} to the pool: "
                    f"{exception.reason}."
                )
                if isinstance(exception.root, usb1.USBErrorBusy):
                    print("Is the `adb` server running? Try `adb kill-server`.")

        if not devices:
            print(f"No devices available to run the session on, aborting.")
        else:
            session = Session(
                devices=devices,
                load=Monkey(
                    100,
                    throttle=1000,
                    allowed_packages=[
                        "com.android.chrome",
                        "com.android.browser",
                    ],
                ),
            )
            session.run()
            self._sessions[str(session.uuid).lower()] = session
